##-------------------------------------------------------------------
##  file name: FourierMethod.R
##  
##  R functions that are used for estimating the central subspace and
##  the central mean subspace.
##
##  need "fourier-clib.dll" (windows) 
##    or "fourier-clib.so" (unix/linux)
##
##  Reference:
##     Zhu and Zeng (2006).
##       Fourier Methods for Estimating the Central Subspace and the
##       Central Mean Subspace in Regression, JASA, 101, 1638--1651.
##  
## Peng Zeng @ Auburn University
## 01/08/2007  
##-------------------------------------------------------------------


FCN <- function(y, x, sw2 = 0.1, st2 = 1.0)
{
   x <- as.matrix(x)
   y <- scale(y)

   n <- nrow(x)
   p <- ncol(x)

   sx.raw <- standardize.x(x)
   sigmahalfinv <- sx.raw$tran.M
   x.std <- sx.raw$x.std

   M <- .C("vecMfcn", as.double(x.std), as.double(y),
           as.integer(n), as.integer(p), as.double(sw2), 
           as.double(st2), M = double(p*p))$M

   M <- matrix(M, ncol=p)
   M.raw <- eigen(M, symmetric = T)
   M.evectors <- sigmahalfinv %*% M.raw$vectors
   M.evectors <- apply(M.evectors, 2, function(x) x/sqrt(sum(x^2)))

   list(evalues = M.raw$values, evectors = M.evectors, M = M)
}


FMN <- function(y, x, sw2 = 0.1)
{
   x <- as.matrix(x)
   y <- scale(y)

   n <- nrow(x)
   p <- ncol(x)

   sx.raw <- standardize.x(x)
   sigmahalfinv <- sx.raw$tran.M
   x.std <- sx.raw$x.std

   M <- .C("vecMfmn", as.double(x.std), as.double(y),
           as.integer(n), as.integer(p), as.double(sw2), 
           M = double(p*p))$M

   M <- matrix(M, ncol=p)
   M.raw <- eigen(M, symmetric = T)
   M.evectors <- sigmahalfinv %*% M.raw$vectors
   M.evectors <- apply(M.evectors, 2, function(x) x/sqrt(sum(x^2)))
   list(evalues = M.raw$values, evectors = M.evectors, M = M)
}


standardize.x <- function(x)
{
   xe <- eigen(cov(x), symmetric = T)
   Mhalf <- xe$vectors %*% diag(1/sqrt(xe$values)) %*% t(xe$vectors)
   x.std <- scale(x, scale=F) %*% Mhalf
   list(x.std = x.std, tran.M = Mhalf)
}

## calculate the distance between the column spaces 
## of two matrix A and B.
## Suppose A and B are orthonormal matrix,
## the distance is the average of eigenvalue (r),
## or the geometry average (q) of matrix B'AA'B.

dist.space <- function(A, B)
{
   A.orth <- qr.Q(qr(A))
   B.orth <- qr.Q(qr(B))

   BAAB <- t(B.orth) %*% A.orth %*% t(A.orth) %*% B.orth
   BAAB.eig <- eigen(BAAB, only.values = T)$values

   list(r = 1 - sqrt(mean(BAAB.eig)), q = 1 - sqrt(prod(BAAB.eig)))
}

##========================== END OF FILE ==========================##
