/*-----------------------------------------------------------------*/
/* file name: fourier-clib.c                                       */
/*                                                                 */
/* C functions needed by FourierMethod.R                           */
/* To complie this file,                                           */
/*       use "Rcmd SHLIB fourier-clib.c" (windows)                 */
/*        or "R CMD SHLIB fourier-clib.c" (unix/linux)             */
/*                                                                 */
/* Reference:                                                      */
/*   Zhu and Zeng (2006).                                          */
/*     Fourier Methods for Estimating the Central Subspace and the */
/*     Central Mean Subspace in Regression, JASA, 101, 1638--1651. */
/*                                                                 */
/* Peng Zeng @ Auburn University                                   */
/* 01/08/2007                                                      */
/*-----------------------------------------------------------------*/

#include <math.h>
#include <R.h>
#include <Rmath.h>

/*=========================== Interface ===========================*/

void vecMfcn(const double* x, const double* y, const int* nn, 
             const int* pp, const double* ssww2, const double* sstt2,
             double* vecM);

void vecMfmn(const double* x, const double* y, const int* nn, 
             const int* pp, const double* ssww2, double* vecM);

/*====================== supported function =======================*/

void Afc(const double* x, const double* y, const int* nn, 
         const int* pp, const double* ssww2, const double* sstt2, 
         double* A);

void Afm(const double* x, const double* y, const int* nn, 
         const int* pp, const double* ssww2, double* A);

void Mnorm(const double* x, const double* A, const int* nn, 
           const int* pp, const double* ssww2, const double* scale, 
           double* vecM);

void IplusMnorm(const double* x, const double* A, const int* nn,
                const int* pp, const double* ssww2, 
                const double* scale, double* vecM);

/*======================== CODE BEGIN HERE ========================*/

/*=================================================================*/
/*      calculate matrix M_{FCn},  output a p-by-p matrix          */
/*=================================================================*/

void vecMfcn(const double* x, const double* y, const int* nn, 
             const int* pp, const double* ssww2, const double* sstt2, 
             double* vecM)
{
   const double one = 1.0;
   double* A;
   A = (double*)Calloc((*nn)*(*nn), double);
   Afc(x, y, nn, pp, ssww2, sstt2, A);
   IplusMnorm(x, A, nn, pp, ssww2, &one, vecM);
   Free(A);
}


/*=================================================================*/
/*      calculate matrix M_{FMn},  output a p-by-p matrix          */
/*=================================================================*/


void vecMfmn(const double* x, const double* y, const int* nn, 
             const int* pp, const double* ssww2, double* vecM)
{
   const double one = 1.0;
   double* A;
   A = (double*)Calloc((*nn)*(*nn), double);
   Afm(x, y, nn, pp, ssww2, A);
   IplusMnorm(x, A, nn, pp, ssww2, &one, vecM);
   Free(A);
}

/*========================== SUBROUTINES ==========================*/

/*=================================================================*/
/* Calculate matrix of type                                        */
/* $$ n^{-2} \sum_{i, j} A_{i, j}                                  */
/*   ((sw2 + 1) x_i - sw2 x_j) ((sw2 + 1) x_j - sw2 x_i)^{\tau} $$ */
/*                                                                 */
/* x is a n-by-p matrix (or n*p vector)                            */
/* A is a n-by-n matrix (or n*n vector)                            */
/* n is the sample size and p is the dimension                     */
/* sw2 is the tuning parameter                                     */
/* scale, times the resulted matrix by scale                       */
/* result are stored in vecM, which is a p-by-p matrix             */
/*                                                                 */
/* for (i, j)th element of vecM                                    */
/* first calculate                                                 */
/*  sum_{ni < nj}                                                  */
/*  A[ni, nj] * (sc1 * (x[ni,i] * x[nj,j] + x[nj,i] * x[ni,j]))    */
/*                                                                 */
/* then plus                                                       */
/*  (sc1 * A[ni, ni] - sc2 * (ni rowsum)) * x[ni,i] * x[ni,j]      */
/*=================================================================*/

void Mnorm(const double* x, const double* A, const int* nn, 
           const int* pp, const double* ssww2, const double* scale, 
           double* vecM)
{
  const int n = (*nn);
  const int p = (*pp);
  const double sw2 = (*ssww2);
  const double sc1 = (sw2 + 1.0) * (sw2 + 1.0) + sw2 * sw2;
  const double sc2 = (sw2 + 1.0) * sw2 * 2.0;
  const double ndiv = (n * n) / (*scale);
  int i, j, ni, nj;
  double* xisc;
  double temp1;

  /* initialize vecM */
  for(i = 0; i < (p*p); i++) vecM[i] = 0.0;

  /* begin to calculate vecM */
  for(i = 0; i < p; i++)
  for(j = i; j < p; j++)
  {
     temp1 = 0.0;
     for(ni = 0; ni < n; ni++)
     for(nj = (ni+1); nj < n; nj++)
     {
        temp1 += ( A[ni + nj*n] *
           (x[ni + i*n] * x[nj + j*n] + x[nj + i*n] * x[ni + j*n]) );
     }

     temp1 *= (sc1 / ndiv);
     vecM[i + j*p] = temp1;
     vecM[j + i*p] = temp1;
  }

  xisc = (double*)Calloc(n, double);
  for(i = 0; i < n; i++)
  {
     temp1 = 0.0;
     for(ni = n*i; ni < n*(i+1); ni++) temp1 += A[ni];
     xisc[i] = sc1 * A[i + i*n] - sc2 * temp1;
  }

  for(i = 0; i < p; i++)
  for(j = i; j < p; j++)
  {
     temp1 = 0.0;
     for(ni = 0; ni < n; ni++)
        temp1 += (xisc[ni] * x[ni + i*n] * x[ni + j*n]);
     vecM[i + j*p] += (temp1 / ndiv);
     vecM[j + i*p] = vecM[i + j*p];
  }

  Free(xisc);
}


/*=================================================================*/
/* calculate the weight matrix                                     */
/* the (i, j)th element is                                         */
/* $exp(-0.5 * st2 (y[i] - y[j])^2 - 0.5 * sw2 * |x[i] - x[j]|^2)$ */
/* the elements on digonal is always 1.0                           */
/*=================================================================*/

void Afc(const double* x, const double* y, const int* nn, 
         const int* pp, const double* ssww2, const double* sstt2, 
         double* A)
{
  const int n = (*nn);
  const int p = (*pp);
  const double sw2 = (*ssww2);
  const double st2 = (*sstt2);
  int i, j, k;
  double temp1, temp2, temp3;

  for(i = 0; i < (n*n); i+=(n+1)) A[i] = 1.0;

  /* calculate the distance matrix A */
  for(i = 0; i < n; i++)
  for(j = (i+1); j < n; j++)
  {
     temp1 = 0.0;
     for(k = 0; k < p; k++)
     {
        temp2 = (x[i + k*n] - x[j + k*n]);
        temp1 += (temp2 * temp2);
     }

     temp2 = (y[i] - y[j]);
     temp3 = exp(-0.5 * (st2 * temp2 * temp2 + sw2 * temp1));
     A[i + j*n] = temp3;
     A[j + i*n] = temp3;
  }
}


/*=================================================================*/
/* calculate the weight matrix                                     */
/* the (i, j)th element is                                         */
/*    $ y[i] * y[j] * exp(- 0.5 * sw2 * |x[i] - x[j]|^2) $         */
/*=================================================================*/

void Afm(const double* x, const double* y, const int* nn, 
         const int* pp, const double* ssww2, double* A)
{
  const int n = (*nn);
  const int p = (*pp);
  const double sw2 = (*ssww2);
  int i, j, k;
  double temp1, temp2, temp3;

  for(i = 0; i < n; i++) A[i + i*n] = y[i] * y[i];

  /* calculate the distance matrix A */
  for(i = 0; i < n; i++)
  for(j = (i+1); j < n; j++)
  {
     temp1 = 0.0;
     for(k = 0; k < p; k++)
     {
        temp2 = (x[i + k*n] - x[j + k*n]);
        temp1 += (temp2 * temp2);
     }

     temp3 = y[i] * y[j] * exp(-0.5 * sw2 * temp1);
     A[i + j*n] = temp3;
     A[j + i*n] = temp3;
  }
}


/*=================================================================*/
/* Calculate matrix of type                                        */
/* $$ n^{-2} \sum_{i, j} A_{i, j} [ sw2 I_p                        */
/*  ((sw2 + 1) x_i - sw2 x_j) ((sw2 + 1) x_j - sw2 x_i)^{\tau} ]$$ */
/*                                                                 */
/* x is a n-by-p matrix (or n*p vector)                            */
/* A is a n-by-n matrix (or n*n vector)                            */
/* n is the sample size and p is the dimension                     */
/* sw2 is the tuning parameter                                     */
/* scale, times the resulted matrix by scale                       */
/* result are stored in vecM, which is a p-by-p matrix             */
/*=================================================================*/

void IplusMnorm(const double* x, const double* A, const int* nn,
                const int* pp, const double* ssww2, 
                const double* scale, double* vecM)
{
   const int n = (*nn);
   const int p = (*pp);

   int i;
   double temp1, Asum;

   Mnorm(x, A, nn, pp, ssww2, scale, vecM);

   Asum = 0.0;
   for(i = 0; i < (n*n); i++) Asum += A[i];
   temp1 = (*scale) * Asum * (*ssww2) / (n * n);
   for(i = 0; i < p; i++) vecM[i + i*p] += temp1;
}

/*========================== END OF FILE ==========================*/
